import importlib
import os
import time
from types import ModuleType
from typing import Tuple, Union

from anytree import Node, exporter

from panoptisch.util import get_file_dir

from panoptisch.imports import (  # isort:skip
    import_file_module,
    import_module,
    resolve_imports,
)


def remove_module_property(node: Node) -> None:
    delattr(node, 'module')
    for child in node.children:
        remove_module_property(child)


def resolve_module(
    module_name: str, parent_file: str, parent_module: ModuleType
) -> Tuple[Union[bool, None], str]:
    '''
    This function resolves a module by imported by the parent_file.
    The resolver behaves like Python's import system.
    If the imported module is a submodule of the parent module,
    the function returns False because:
    1. It is a nested module, not a dependency of the parent.
    2. The files from the submodule as included in the parent module,
    so it's imports will be parsed and factored in as the parent's imports.
    TODO: handle None
    '''
    if module_name != 'None':
        parent_dir = get_file_dir(parent_file)
        is_part_of_module = parent_file.endswith(
            '__init__.py'
        ) or os.path.exists(f'{parent_dir}/__init__.py')
        module_as_dir = f'{parent_dir}/{module_name}'
        module_as_file = f'{module_as_dir}.py'
        if os.path.isfile(module_as_file):
            if not is_part_of_module:
                module = import_file_module(module_name, module_as_file)
                return module, module_name
            else:
                return False, module_name
        elif os.path.isdir(module_as_dir):
            if not is_part_of_module:
                # this check matters as a simple script cannot
                # import another script in a folder if the folder does not have
                # an __init__.py (making it a module)
                if os.path.exists(f'{module_as_dir}/__init__.py'):
                    module = import_file_module(
                        module_name, f'{module_as_dir}/__init__.py'
                    )
                    return module, module_name
            else:
                return False, module_name
        try:
            return import_module(module_name), module_name
        except ModuleNotFoundError:
            print(
                f'AAAAA Could not resolve: {module_name}, parent = {parent_module.__name__}'  # noqa E501
            )
    return None, module_name


def get_imports(
    current_root: Node,
    stdlib_dir=None,
    depth=0,
    max_depth=15,
    omit_not_found=False,
):
    depth += 1
    if depth == max_depth:
        return
    imports = resolve_imports(current_root.module)
    for item in imports:
        file = item.get('file')
        file_imports = item.get('imports')
        for module_name in file_imports:
            if module_name != current_root.module.__name__:
                if (
                    stdlib_dir
                    and stdlib_dir in file
                    and 'site-packages' not in file
                ):
                    continue
                module, name = resolve_module(
                    module_name, file, current_root.module
                )
                if module:
                    node = Node(
                        name,
                        module=module,
                        parent=current_root,
                    )
                    get_imports(
                        node,
                        stdlib_dir=stdlib_dir,
                        depth=depth,
                        max_depth=max_depth,
                    )
                else:
                    if module is not False and omit_not_found:
                        node = Node(
                            module_name,
                            parent=current_root,
                            module=None,
                            reason='ModuleNotFoundError',
                        )


def run(args):
    print('-' * 50)
    print(f'Running Panopticon on module {args.module}')
    print('-' * 50)
    print(f'maximum dependency depth: {args.max_depth}')
    print(f'stdlib dir is {args.stdlib_dir}')
    print(f'omit not found: {args.omit_not_found}')
    print('-' * 50)
    start = time.time()
    if args.module.endswith('.py'):
        spec = importlib.util.spec_from_file_location(
            args.module.replace('.py', ''), args.module
        )
        root_module = importlib.util.module_from_spec(spec)
    else:
        root_module = import_module(args.module)
    root_node = Node(root_module.__name__, module=root_module)
    get_imports(
        root_node,
        max_depth=args.max_depth,
        stdlib_dir=args.stdlib_dir,
        omit_not_found=args.omit_not_found,
    )
    remove_module_property(root_node)
    with open(f'{args.out}', 'w') as f:
        f.write(exporter.JsonExporter(indent=4).export(root_node))
    end = time.time()
    elapsed = '%.2fs' % (end - start)
    print('-' * 50)
    print(f'Done in {elapsed}')
    print(f'See {args.out}')



import ast
import glob
import importlib
from types import ModuleType
from typing import Dict, List, Literal, Union

from panoptisch.util import get_file_dir
from panoptisch.visitor import Visitor


def get_module_files(module: ModuleType) -> List[str]:
    try:
        entry = module.__file__
    except AttributeError:
        return []  # frozen module
    if entry.endswith('__init__.py'):
        entry_folder = get_file_dir(entry)
        return glob.glob(f'{entry_folder}/**/*.py', recursive=True)
    else:
        return [entry]


def get_ast_from_file(filename: str) -> ast.Module:
    with open(filename) as f:
        try:
            return ast.parse(f.read())
        # ast module has test files that are bad syntax
        # to test unicode & syntax parsing
        # so the errors are ignored.
        # TODO: ignore test files
        except UnicodeDecodeError:
            return ast.Module()
        except SyntaxError:
            return ast.Module()


RESOLVED_IMPORT_LIST = List[
    Dict[Union[Literal['file'], Literal['imports']], Union[str, List[str]]]
]


def resolve_imports(module: ModuleType) -> RESOLVED_IMPORT_LIST:
    visitor = Visitor()
    files = get_module_files(module)
    imports: RESOLVED_IMPORT_LIST = []
    for file in files:
        if not file.endswith('.so'):  # c extension
            file_ast = get_ast_from_file(file)
            visitor.visit(file_ast)
            imports.append({'file': file, 'imports': visitor.imports})
            visitor.flush_imports()
    imports = simplify_imports(imports)
    return imports


def simplify_imports(imports: RESOLVED_IMPORT_LIST) -> RESOLVED_IMPORT_LIST:
    simplified = []
    existing = []
    for item in imports:
        copy = item.copy()
        root_modules = map(
            lambda x: x if len(x.split('.')) == 0 else x.split('.')[0],
            item['imports'],
        )
        copy['imports'] = list(
            set(filter(lambda x: x not in existing, root_modules))
        )
        simplified.append(copy)
        existing.extend(copy['imports'])
    return simplified


def import_module(name: str) -> ModuleType:
    return importlib.import_module(name)


def import_file_module(name: str, path: str) -> ModuleType:
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    return module



import ast
from typing import List


class Visitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.imports: List[str] = []

    def visit_Import(self, node: ast.Import):
        self.imports.extend([alias.name for alias in node.names])

    def visit_ImportFrom(self, node: ast.ImportFrom):
        module_from = node.module
        self.imports.extend(
            [f'{module_from}.{alias.name}' for alias in node.names]
        )

    def flush_imports(self):
        self.imports = []



import argparse

from panoptisch.lib_resolver import get_stdlib_dir
from panoptisch.scanner import run


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'module',
        action='store',
        help='Name of module or file you wish to scan.',
    )
    parser.add_argument(
        '--show-stdlib-dir',
        action='store_true',
        help='Prints the automatically resolved stdlib directory.',
    )
    parser.add_argument(
        '--max-depth',
        type=int,
        action='store',
        help='Maximum dependency depth.',
        default=3,
    )
    parser.add_argument(
        '--out',
        action='store',
        help='File to output report.',
        default='out.json',
    )
    parser.add_argument(
        '--auto-stdlib-dir',
        action='store_true',
        help='Ignore stdlib modules by automatically resolving their path. MAY BE BUGGY. Try running panoptisch <module_name> --show-stdlib-dir to see the directory before using this.',  # noqa E501
    )
    parser.add_argument(
        '--stdlib-dir',
        type=str,
        action='store',
        help='Ignore stdlib modules by providing their path',
    )
    parser.add_argument(
        '--omit-not-found',
        action='store_true',
        help='Do not include modules that could not be resolved in report',
    )
    args = parser.parse_args()
    if args.show_stdlib_dir:
        print(get_stdlib_dir())
    else:
        if args.auto_stdlib_dir:
            args.stdlib_dir = get_stdlib_dir()
        run(args)


__all__ = ['main']



import sys


def get_stdlib_dir():
    major = sys.version_info.major
    minor = sys.version_info.minor
    if sys.platform not in ['win32', 'cygwin']:
        for path in sys.path:
            if path.endswith(f'/lib/python{major}.{minor}'):
                return path
    raise Exception(
        "Panopticon cannot automatically resolve Python's stdlib directory. Please provide it yourself with --stdlib-dir."
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['panoptisch']

package_data = \
{'': ['*']}

install_requires = \
['anytree>=2.8.0,<3.0.0']

entry_points = \
{'console_scripts': ['panoptisch = panoptisch:main']}

setup_kwargs = {
    'name': 'panoptisch',
    'version': '0.1.1',
    'description': 'A recursive Python dependency scanner.',
    'long_description': " ## Panoptisch: A recursive dependency scanner for Python projects\n\n####  ⚠️🚨 Early stage! May not work as expected 🚨⚠️\n#### What?\nPanoptisch scans your Python file or module to find it's imports (aka dependencies) and recursively does so for all dependencies and sub-dependencies.\nIt then generates a dependency tree in JSON for you to parse and enforce import policies.\nImports are resolved by mimicing Python's import system. It's completely static besides the importing of modules to find the location of its source file(s).\n\n##### Please NOTE:\nThere are known **limitations and issues** at this stage. Please read this before using Panoptisch.  \nSee: ``LIMITATIONS.md`` [LINK](LIMITATIONS.md).\n\n\n#### Motivation\nI was not able to find a proper dependency scanner for Python. Panoptisch was born out of the need to accurately verify dependency usage accross an entire project.  \nIt's aim is to generate a JSON report that can be parsed and evaluated to **assert import policies**.  \nFor example, you may want to restrict ``os``, ``socket``, ``sys`` and ``importlib`` imports to selected packages.\n\n\n#### Usage\n\n1. Install ``Panoptisch`` in the same virtual environment as your project, this is important!  \n```\npip install panoptisch\n```\n\n2. Use\n```\nusage: panoptisch <module>\n\npositional arguments:\n  module                Name of module or file you wish to scan.\n\noptions:\n  -h, --help            show this help message and exit.\n  --show-stdlib-dir     Prints the automatically resolved stdlib directory.\n  --max-depth MAX_DEPTH\n                        Maximum dependency depth.\n  --out OUT             File to output report.\n  --auto-stdlib-dir     Ignore stdlib modules by automatically resolving their path. MAY BE BUGGY. Try running panoptisch <module_name> --show-stdlib-dir to see the directory before using this.\n  --stdlib-dir STDLIB_DIR Ignore stdlib modules by providing their path.\n  --omit-not-found      Do not include modules that could not be resolved in report.\n```\nA typical run may be\n```\n$ panoptisch <module or file> --max-depth 5 --omit-not-found\n```\n3. See report\n```\n$ more out.json\n```\n#### LICENSE\nAll work is licensed under the [GNU General Public License Version 3](https://www.gnu.org/licenses/gpl-3.0.en.html).\n\n#### Contributing\nFeedback, contributions and issues welcome. \n\n",
    'author': 'aarnav',
    'author_email': 'aarnavbos@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/R9295/panoptisch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
